/* "$Id: Calendar.c,v 1.22 2005/06/09 10:01:45 jjf Exp $" */

/*
  This file is part of xdeskcal.
  Copyright  2005 J Farlow, jjf1125@earthlink.net.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

# include <X11/Xlib.h>
# include <X11/Xos.h>
# include <X11/IntrinsicP.h>
# include <X11/StringDefs.h>
# include <X11/extensions/shape.h>
# include <X11/Xaw/XawInit.h>
# include <X11/Xmu/Converters.h>
# include <stdio.h>
# include <stdlib.h>
# include "CalendarP.h"

extern WidgetClass calendarWidgetClass;

typedef struct _CalendarRec *CalendarWidget;
typedef struct _CalendarClassRec *CalendarWidgetClass;

#define offset(field) XtOffsetOf(CalendarRec, calendar.field)

#define XtNshapeWindow	"shapeWindow"
#define XtCShapeWindow	"ShapeWindow"
#define XtNmonth        "month"
#define XtCMonth        "Month"
#define XtNyear         "year"
#define XtCYear         "Year"
#define XtNnoHighlight  "nohighlight"
#define XtCNoHighlight  "Nohighlight"

XtResource resources[] = {
    {XtNfont, XtCFont, XtRString, sizeof(char *), offset(font), 
     XtRString, "fixed"},
    {XtNshapeWindow, XtCShapeWindow, XtRBoolean, sizeof (Boolean),
     offset (shape_window), XtRImmediate, (XtPointer) TRUE},
    {XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
     offset(fgp), XtRString, "white"},
    {XtNbackground, XtCBackground, XtRPixel, sizeof(Pixel),
     offset(bgp), XtRString, "black"},
    {XtNmonth, XtCMonth, XtRString, sizeof (int),offset (month), 
     XtRImmediate, NULL},
    {XtNyear, XtCYear, XtRInt, sizeof (int),offset (year), XtRImmediate, NULL},
    {XtNnoHighlight, XtCNoHighlight, XtRInt, sizeof (int),
     offset(nohighlight), XtRImmediate, (XtPointer) False},
};

static void ClassInitialize (), Initialize(), Realize(), Redisplay();

int mon_1st_wday (int, int);
int year_1st_wday (int);
void days (CalendarWidget);
void draw (CalendarWidget, GC, Pixmap);
void update (XtPointer, XtIntervalId *);
int wdaytitle (CalendarWidget);
int space_width (CalendarWidget);
static struct tm *timenow (void);

CalendarClassRec calendarClassRec = {
    { /* core fields */
      /* superclass		*/	&widgetClassRec,
      /* class_name		*/	"Calendar",
      /* size			*/	sizeof(CalendarRec),
      /* class_initialize	*/	ClassInitialize,
      /* class_part_initialize	*/	NULL,
      /* class_inited		*/	FALSE,
      /* initialize		*/	Initialize,
      /* initialize_hook	*/	NULL,
      /* realize		*/	Realize,
      /* actions		*/	NULL,
      /* num_actions		*/	0,
      /* resources		*/	resources,
      /* num_resources		*/      XtNumber(resources),
      /* xrm_class		*/	NULLQUARK,
      /* compress_motion	*/	TRUE,
      /* compress_exposure	*/	TRUE,
      /* compress_enterleave	*/	TRUE,
      /* visible_interest	*/	FALSE,
      /* destroy		*/	NULL,
      /* resize			*/	NULL,
      /* expose			*/	Redisplay,
      /* set_values		*/	NULL,
      /* set_values_hook	*/	NULL,
      /* set_values_almost	*/	NULL,
      /* get_values_hook	*/	NULL,
      /* accept_focus		*/	NULL,
      /* version		*/	XtVersion,
      /* callback_private	*/	NULL,
      /* tm_table		*/	NULL,
      /* query_geometry		*/	XtInheritQueryGeometry,
    }
};

int month_lengths[] = {0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
char *monthnames[13] = {"", "January", "February", "March", "April", 
			"May", "June", "July", "August", "September",
		       "October", "November", "December"};
char *cal_d_names_fixed[8] = {"Su","Mo","Tu","We","Th","Fr","Sa","Su"};
/* "Tu" label needs space hint for non-monospaced fonts. */
char *cal_d_names_prop[8] = {"Su","Mo","Tu ","We","Th","Fr","Sa","Su"};
char **cal_d_names;			

/* Size hint for text border. */
#define PADDING 5

/* Maximum width and height in number of characters. */
#define TEXT_CAL_WIDTH 20
#define TEXT_CAL_HEIGHT 9

#define PROPORTIONAL_FONT(fstructptr) \
         ((fstructptr)->min_bounds.width != \
         (fstructptr)->max_bounds.width) ? 1 : 0
#define FONT_HEIGHT(w) ((w)->calendar.font_struct->ascent + \
                       (w)->calendar.font_struct->descent)

#define YEAR(w) (w)->calendar.year
#define MONTH(w) (w)->calendar.month
#define DATE(w) (w)->calendar.date

static void ClassInitialize () {
    XtAddConverter( XtRString, XtRBackingStore, XmuCvtStringToBackingStore,
		    NULL, 0 );
}

static void Initialize(Widget greg, Widget gnew, ArgList args, 
		       Cardinal *n_args) {

  CalendarWidget w = (CalendarWidget)gnew;
  int shape_event_base, shape_error_base;
  struct tm *now;
  char **fonts;
  int n_fonts;

  if (!MONTH(w) && !YEAR(w)) w->calendar.update = True;

  now = timenow ();

  MONTH(w) = MONTH(w) ? MONTH(w) : now->tm_mon + 1;
  YEAR(w) = YEAR(w) ? YEAR(w) : now->tm_year + 1900;
  DATE(w) = now->tm_mday;
  
  fonts = (char **)XListFonts (XtDisplay (w), w->calendar.font, 10, &n_fonts);

  if (!n_fonts) {
    fprintf (stderr, "xdeskcal: Font %s does not exist, using fixed font.\n",
	     w->calendar.font);
    strcpy (w->calendar.font, "fixed");
  }

  w->calendar.font_info = XLoadFont (XtDisplay(w), w->calendar.font);
  w->calendar.font_struct = XQueryFont (XtDisplay(w), w->calendar.font_info);

  if (PROPORTIONAL_FONT(w->calendar.font_struct))
    cal_d_names = cal_d_names_prop;
  else
    cal_d_names = cal_d_names_fixed;

  w->core.width = wdaytitle (w);
                                  /* Extra space for label hints. */
  w->core.width += (PADDING * 2) + PADDING;

  w->core.height = FONT_HEIGHT(w) * TEXT_CAL_HEIGHT + (PADDING * 2);

  if (w->calendar.shape_window && 
      !XShapeQueryExtension (XtDisplay (w), &shape_event_base,
			     &shape_error_base))
    w->calendar.shape_window = False;
}

static void Realize(Widget gw, XtValueMask *valueMask,
		     XSetWindowAttributes *attrs) {

  CalendarWidget w = (CalendarWidget)gw;

  *valueMask |= CWBackingStore;
  attrs-> backing_store = Always;

  XtCreateWindow( gw, (unsigned)InputOutput, (Visual *)CopyFromParent,
		  *valueMask, attrs );

  if (w->calendar.shape_window)
    XSetWindowBackground (XtDisplay(w), XtWindow(w), w->calendar.fgp);
  else
    XSetWindowBackground (XtDisplay(w), XtWindow(w), w->calendar.bgp);

  days (w);
  Redisplay (gw);

  if (w->calendar.update)
    XtAppAddTimeOut (XtWidgetToApplicationContext((Widget) w), 
		     30 * 1000, update, (XtPointer)w);
}

static void Redisplay(Widget gw) {
  
  CalendarWidget w = (CalendarWidget) gw;
  GC gc;
  Pixmap pixmap;
  XGCValues	xgcv;
  int x, y;
  Widget parent;

  pixmap = XCreatePixmap (XtDisplay (w),
		  XtWindow (w), w->core.width, w->core.height, 
    (w->calendar.shape_window ? 1: 
     DefaultDepth (XtDisplay (w), DefaultScreen (XtDisplay (w)))));

  gc = XCreateGC (XtDisplay (w), pixmap, 0, &xgcv);

  if (w->calendar.shape_window) {
    XSetForeground (XtDisplay (w), gc, 0);
    XFillRectangle (XtDisplay (w), pixmap, gc,
		    0, 0, w->core.width, w->core.height);
    XSetForeground (XtDisplay (w), gc, 1);
  } else {
    XSetForeground (XtDisplay (w), gc, w->calendar.bgp);
    XFillRectangle (XtDisplay (w), pixmap, gc,
		    0, 0, w->core.width, w->core.height);
    XSetForeground (XtDisplay (w), gc, w->calendar.fgp);
  }

  XSetFont (XtDisplay(w), gc, w->calendar.font_info);

  XSetLineAttributes (XtDisplay(w), gc, 2, LineSolid, CapButt, JoinBevel);

  draw (w, gc, pixmap);

  if (w->calendar.shape_window) {
    x = y = 0;
    for (parent = (Widget) w; XtParent (parent); parent = XtParent (parent)) {
      x += parent->core.x + parent->core.border_width;
      x += parent->core.y + parent->core.border_width;
    }
    XShapeCombineMask (XtDisplay (parent), XtWindow (parent),
		       ShapeBounding,
		       x, y, pixmap,
		       ShapeSet);
  } else {
    XCopyArea (XtDisplay (w), pixmap, 
	       XtWindow (w), gc, 
	       0, 0, w->core.width, w->core.height, 
	       0, 0);
  }

  XFreePixmap (XtDisplay (w), pixmap);
  XtReleaseGC (gw, gc);

}

void update (XtPointer wp, XtIntervalId *id) {

    CalendarWidget w = (CalendarWidget)wp;
    struct tm *now;

    now = timenow ();

    YEAR(w) = ((YEAR(w) != now->tm_year + 1900) ? 
	       now->tm_year + 1900 : YEAR(w));

    if (MONTH(w) != now->tm_mon + 1) {
      MONTH(w) = now->tm_mon + 1;
      days (w);
    }

    if (now->tm_mday != DATE(w)) {
      DATE(w) = now->tm_mday;
      XClearWindow (XtDisplay (w), XtWindow (w));
      Redisplay ((Widget)w);
    }

    XtAppAddTimeOut (XtWidgetToApplicationContext((Widget) w), 
		     30 * 1000, update, w);
}

void days (CalendarWidget w) {
  
  int i, row, d, first, direction, ascent, descent;
  char line[16];
  XCharStruct char_box;

  first = mon_1st_wday (w->calendar.month, w->calendar.year);
  i = row = 0; 
  for (d = 1; d <= month_lengths[w->calendar.month];) {
    if (i < (first - 1) && (row == 0)) {
      w->calendar.days[row][i].day_no = 
	w->calendar.days[row][i].month_no = 
	w->calendar.days[row][i].year_no = 0;
    } else {
      sprintf (line, "%2d", d);
      w->calendar.days[row][i].day_no = d++;
      w->calendar.days[row][i].month_no = w->calendar.month;
      w->calendar.days[row][i].year_no = w->calendar.year;

      XTextExtents (w->calendar.font_struct, line, strlen(line), 
		    &direction, &ascent, &descent, &char_box);
      
      w->calendar.days[row][i].width =     
	(char_box.rbearing - char_box.width) + char_box.lbearing + 
	char_box.width;
      w->calendar.days[row][i].height = FONT_HEIGHT(w);
    }
    (void)i++;
    if (i && (i % 7) == 0) {
      row++;
      i = 0;
    }
  }
}

void draw (CalendarWidget w, GC gc, Pixmap pixmap) {

  char line[TEXT_CAL_WIDTH+1];
  XCharStruct b_box;
  int i, row, d, line_y_org, char_x_org, w_space;
  int direction, ascent, descent;
  struct tm *now;

  now = timenow ();

  if (w->calendar.year < 2000) 
    w->calendar.year += 2000;

  sprintf (line, "%s %d", monthnames[w->calendar.month], w->calendar.year);

  XTextExtents (w->calendar.font_struct, line, strlen(line), 
		&direction, &ascent, &descent, &b_box);
  
  line_y_org = PADDING + FONT_HEIGHT(w);
  char_x_org = ((w->core.width - PADDING) - b_box.width) / 2;
  
  XDrawString (XtDisplay(w), pixmap, gc,
	       char_x_org,
	       line_y_org,
	       line, strlen(line));

  line_y_org += PADDING + FONT_HEIGHT(w);
  char_x_org = PADDING;

  w_space = space_width (w);
  for (i = 0; i <= 6; i++) {
    XTextExtents (w->calendar.font_struct, 
		  w->calendar.wdays[i].name,strlen(w->calendar.wdays[i].name),
		  &direction, &ascent, &descent, &b_box);
    XDrawString (XtDisplay(w), pixmap, gc, 
		 /* Align in column. */
		 char_x_org + w->calendar.wdays[i].width - 
		   (b_box.rbearing - b_box.lbearing),
		 line_y_org,
		 w->calendar.wdays[i].name, strlen(w->calendar.wdays[i].name));

    w->calendar.wdays[i].x = char_x_org;
    w->calendar.wdays[i].y = line_y_org;
    char_x_org += w->calendar.wdays[i].width + w_space;
  }

  line_y_org += PADDING + FONT_HEIGHT(w);

  i = row = 0;
  for (d = 1; d<= month_lengths[w->calendar.month];) {
    if (w->calendar.days[row][i].day_no) {
      sprintf (line, "%2d", w->calendar.days[row][i].day_no);
      XTextExtents (w->calendar.font_struct, line, strlen(line), 
		    &direction, &ascent, &descent, &b_box);
      /* Align in column. */
      char_x_org = w->calendar.wdays[i].x + 
	(w->calendar.wdays[i].width - b_box.rbearing - b_box.lbearing);
  
      XDrawString (XtDisplay(w), pixmap, gc,
		   char_x_org,
		   line_y_org,
		   line, strlen(line));

      if (! w->calendar.nohighlight) 
	if ((d == now->tm_mday) && (MONTH(w) == now->tm_mon + 1) &&
	    (YEAR(w) == now->tm_year + 1900))
	  {
	    XDrawLine (XtDisplay (w), pixmap, gc,
		       char_x_org,
		       line_y_org + 2,
		       char_x_org + (w->calendar.days[row][i].width),
		       line_y_org + 2);

	  }
      (void)d++;
    }
      
    
    if (i &&  (i % 6) == 0) {
      (void)row++;
      line_y_org += FONT_HEIGHT(w);
      i = 0;
    } else {
      (void)i++;
    }
  }
}

int space_width (CalendarWidget w) {

  XCharStruct sp_box;
  int direction, ascent, descent;
  XTextExtents (w->calendar.font_struct, " ", strlen(" "),  
		&direction, &ascent, &descent, &sp_box);
  return sp_box.width;
}

/* Return the width of the title. */
int wdaytitle (CalendarWidget w) {

  int i, direction, descent, ascent, max_label_width = 0 ;
  XCharStruct bbox;
  int w_space;

  w_space = space_width (w);

  for (i = 0; i <= 6; i++) {
    strcpy (w->calendar.wdays[i].name, cal_d_names[i]);
    XTextExtents (w->calendar.font_struct, 
		  cal_d_names[i], strlen(cal_d_names[i]), 
		  &direction, &ascent, &descent, &bbox);
    w->calendar.wdays[i].height = FONT_HEIGHT(w);
    max_label_width = 
      (bbox.width > max_label_width) ? bbox.width : max_label_width;
  }

  for (i = 0; i <= 6; i++) 
    w->calendar.wdays[i].width = max_label_width;

  return (max_label_width * 7) + (w_space * 6);
}

/* returns day of week of Jan1 of given year, 1-7 */
int year_1st_wday (int year) {

  int day;

  day = 4 + year + (year + 3)/4;
  if (year > 1800) {
    day -= (year - 1701) / 100;
    day += (year - 1601) / 400;
  }
  if (year > 1752) {
    day += 3;
  }
  return (day % 7);
}

int mon_1st_wday (int month, int year) {

  int day, x, i;
  
  day = year_1st_wday (year);
  x = ((year_1st_wday (year + 1) + 7 - day) % 7);
  if (x == 1) {
    month_lengths[2] = 28;
  } else {
    if (x != 2) {
      month_lengths[9] = 19;
    }
  }
  for (i = 1; i < month; ++i) {
    day += month_lengths[i];
  }
  
  day = day % 7;
  if (day > 0) {
  } else {
    day = 7;
  }
  day = day + 1 - 7;
  if (day > 0) {
    return (day);
  } else {
    return (7 + day);
  }

}

static struct tm *timenow (void) {

  struct tm *now;
  time_t utctime;

  utctime = time (NULL);
  now = localtime (&utctime);
  return now;
}
